$ErrorActionPreference = "Stop"

function exitIfDeployFailed() {
  if ($Script:deployFailed -eq $true) {
    throw 'Exiting due to failed deployment.'
  }
}

function execute_on_build_node([string] $command_name, [string] $command) {
  if (-not $command_name -or -not $command) {
    throw "Usage: execute_on_build_node command_name command"
  }

  $commandToRun = $command
  if ($dry_run -eq $true) {
    # Echo a message on the build node instead of running the deployCommand.
    $dryRunMessage = $DRY_RUN_MESSAGE -f $command_name
    $commandToRun = "Write-Output ${dryRunMessage}"
  }
  Write-Output "Running command: `'${command_name}`' | On build node."
  Invoke-Expression $commandToRun
}

function execute_on_target_cluster([string] $command_name, [string] $command) {
  if (-not $command_name -or -not $command) {
    throw "Usage: execute_on_target_cluster command_name command"
  }
  $sshOptions = "${baseSSHOptions} -i ~/.ssh/${Env:target_cluster_name}"
  return execute_on_cluster $Env:target_cluster_name $command_name $command
}

function execute_on_current_cluster([string] $command_name, [string] $command) {
  if (-not $command_name -or -not $command) {
    throw "Usage: execute_on_current_cluster command_name command"
  }
  $sshOptions = "${baseSSHOptions} -i ~/.ssh/${Env:current_cluster_name}"
  return execute_on_cluster $Env:current_cluster_name $command_name $command
}

function execute_on_cluster([string] $vm_cluster_resource_name, [string] $command_name, [string] $command) {
  if (-not $vm_cluster_resource_name -or -not $command_name -or -not $command) {
    throw "Usage: execute_on_cluster vm_cluster_resource_name command_name command"
  }

  $vmTargets = (Get-Variable -Name "res_$( $vm_cluster_resource_name )_targets").Value | ConvertFrom-Json

  if ($null -ne $deploy_targets_override -and $strategy -ne $BLUE_GREEN_STRATEGY) {
    Write-Output "Overriding vm deploy targets with: ${deploy_targets_override}"
    $vmTargets = $deploy_targets_override.Split(",")
  }

  execute_command "Write-Output `"Running command: `'${command_name}`' | On cluster: `'${vm_cluster_resource_name}`'`""

  for ($i = 0; $i -lt $vmTargets.Length; $i++) {
    $vmTarget = Invoke-Expression $vmTargets[$i]
    # Wait between deploys if delay was specified
    if ($rolloutDelayCommandConfigVal -ne "" -and $i -ne 0) {
      Start-Sleep -Seconds $rolloutDelayCommandConfigVal
    }
    $sshBaseCmd = "ssh ${sshOptions} -p ${SSH_PORT} ${sshUserConfigVal}@${vmTarget}"

    # Command to source the file with vmEnvironmentVariables
    if ($vmEnvVarsLen -ne "") {
      # Remove the env file right after sourcing. There could be secrets in there.
      $sourceEnvFile = "source ${Script:targetDir}/${vmEnvFilename};"
    }

    $fullCommand = "${sshBaseCmd} `"${sourceEnvFile} `${command}`""
    if ($dry_run -eq $true) {
      # Echo a message on the vm instead of running the deployCommand.
      $dryRunMessage = $DRY_RUN_MESSAGE -f $command_name
      $fullCommand = "${sshBaseCmd} `"echo ${dryRunMessage}`""
    }

    try {
      Write-Output "Running command: `'${command_name}`' | On vm: `'${vmTarget}`'"
      Write-Output "--- Command ---"
      Write-Output "Name: ${command_name}`n" "VM Address: ${vmTarget}`n" "Command: `"${command}`""
      Write-Output "--- Command Output ---"
      Invoke-Expression $fullCommand
      if ($lastexitcode) {
        throw
      }

      # rollback archives will be created on deployed_vms with rollingUpdate strategy.
      # It does not matter for blueGreen.
      $Script:deployedVMs += $vmTarget
    }
    catch {
      $Script:deployFailed = $true
      # Don't exit on failed commands if fastFail is specified as false
      if ($fastFailConfigVal -eq $false) {
        continue
      }
      break
    }
  }
  exitIfDeployFailed
}

function DeployApplication() {
  # export current_command for use in onFailure.
  $Env:current_command = ""

  $BLUE_GREEN_STRATEGY = "blueGreen"
  $DRY_RUN_MESSAGE = 'dry_run is true. "{0}" will not run.'
  $DEFAULT_SCRIPT_NAME = "anonymous"
  $SCRIPT_CONTEXT_BUILD_NODE = "buildNode"
  $SCRIPT_CONTEXT_TARGET_CLUSTER = "targetCluster"
  $SCRIPT_CONTEXT_CURRENT_CLUSTER = "currentCluster"
  $SSH_PORT = "22"

  # Without these options, ssh will hang forever in the event of an error
  $baseSSHOptions = "-o BatchMode=yes -o StrictHostKeyChecking=no"
  $strategy = find_step_configuration_value 'strategy'

  if ($strategy -eq "") {
    execute_command "Write-Output Defaulting to rollingUpdate strategy."
    $strategy = "rollingUpdate"
  }

  $vmClusterResNames = @(get_resource_names -type VmCluster -operation "IN")

  if ($vmClusterResNames -eq "") {
    throw "A VmCluster input resource is required."
  }

  # validate config related to blueGreen strategy
  if ($strategy -eq $BLUE_GREEN_STRATEGY -and $vmClusterResNames.Length -ne 2 ) {
    throw "Exactly two VmCluster resources are required when using blueGreen strategy."
  }

  $buildinfoResName = get_resource_name -type BuildInfo -operation "IN"
  $filespecResName = get_resource_name -type FileSpec -operation "IN"
  $releaseBundleResName = get_resource_name -type ReleaseBundle -operation "IN"

  $deployableResources = @($buildinfoResName, $filespecResName, $releaseBundleResName).Where({ $_.Length })

  if (@($deployableResources).Length -ne 1) {
    throw "Exactly one resource of type BuildInfo`|FileSpec`|ReleaseBundle is supported."
  }

  # We put everything we want to upload to vms in a directory
  # We will create a tarball from all of it
  $tardir = Join-Path $PWD -ChildPath "deploy-artifacts"
  execute_command "mkdir ${tardir}"

  # Create a file with env vars to source on the target vms
  $vmEnvFilename = "${step_name}_${run_id}.env"
  $tmpVMEnvFilePath = "${tardir}\tmp_${vmEnvFilename}"
  $vmEnvFilePath = "${tardir}\${vmEnvFilename}"
  $vmEnvVarsLen = find_step_configuration_value("vmEnvironmentVariables_len")
  if ($vmEnvVarsLen -ne "") {
    for ($i = 0; $i -lt $vmEnvVarsLen; $i++) {
      $envVar = find_step_configuration_value("vmEnvironmentVariables_$i")
      Add-Content -Path $tmpVMEnvFilePath -Value "export ${envVar}"
    }

    # Make file unix compatible
    $content = ((Get-Content -Raw $tmpVMEnvFilePath) -replace "`r`n","`n")
    [io.file]::WriteAllText($vmEnvFilePath, $content)
  }

  Push-Location $tardir
  if ($buildinfoResName -ne "") {
    try {
      $buildinfoNumber = (Get-Variable -Name "res_$( $buildinfoResName )_buildNumber").Value
      $buildinfoName = (Get-Variable -Name "res_$( $buildinfoResName )_buildName").Value
    } catch {
      throw "buildName and buildNumber are required to use a BuildInfo resource with this step."
    }
    $buildinfoURL = (Get-Variable -Name "res_$( $buildinfoResName )_sourceArtifactory_url").Value
    $buildinfoUser = (Get-Variable -Name "res_$( $buildinfoResName )_sourceArtifactory_user").Value
    $buildinfoApikey = (Get-Variable -Name "res_$( $buildinfoResName )_sourceArtifactory_apikey").Value
    execute_command "retry_command ${jfrog_cli_path} rt download `"*`" ${tardir}\ --build=${buildinfoName}/${buildinfoNumber} --url=${buildinfoURL} --user=${buildinfoUser} --password=`${buildinfoApikey} --insecure-tls=${no_verify_ssl}"
  }
  elseif ($filespecResName -ne "") {
    $filespecResPath = (Get-Variable -Name "res_$( $filespecResName )_resourcePath").Value
    execute_command "mv $filespecResPath\* $tardir"
  } elseif ($releaseBundleResName -ne "") {
    $rtIntName = get_integration_name -type Artifactory
    if ($releaseBundleResName -eq "") {
      throw "An Artifactory integration is required to use a ReleaseBundle resource with LinuxVmDeploy steps."
    }
    $releaseBundleURL = (Get-Variable -Name "int_$( $rtIntName )_url").Value
    $releaseBundleUser = (Get-Variable -Name "int_$( $rtIntName )_user").Value
    $releaseBundleApikey = (Get-Variable -Name "int_$( $rtIntName )_apikey").Value
    $releaseBundleName = (Get-Variable -Name "res_$( $releaseBundleResName )_name").Value
    $releaseBundleVersion = (Get-Variable -Name "res_$( $releaseBundleResName )_version").Value
    execute_command "retry_command ${jfrog_cli_path} rt download `"*`" ${tardir}\ --bundle=${releaseBundleName}/${releaseBundleVersion} --url=${releaseBundleURL} --user=${releaseBundleUser} --password=`${releaseBundleApikey} --insecure-tls=${no_verify_ssl}"
  }
  $tarballName = "${pipeline_name}-${run_id}.tar.gz"
  execute_command "tar -czvf ../${tarballName} ."
  Pop-Location

  $sshUserConfigVal = find_step_configuration_value("sshUser")
  $rolloutDelayCommandConfigVal = find_step_configuration_value("rolloutDelaySecs")
  $fastFailConfigVal = find_step_configuration_value("fastFail")
  $Script:targetDir = "/home/${sshUserConfigVal}/${step_name}/${run_id}"
  $targetDirConfigVal = find_step_configuration_value("targetDirectory")
  if ($targetDirConfigVal -ne "") {
    $Script:targetDir = $targetDirConfigVal
  }

  # export name of target cluster for use in commands defined in step configuration (e.g. deployCommand).
  $Env:target_cluster_name = $vmClusterResNames[0]

  # Used for blueGreen strategy only. The cluster that was live at the beginning of the  step.
  $Env:current_cluster_name = ""

  # For blueGreen, try to determine from step state which cluster to deploy to.
  # If we have nothing in state, just use the first one.
  $lastDeployedClusterFilename = "last_deployed_cluster"
  $stepStateDir = "${step_workspace_dir}"
  if ($strategy -eq $BLUE_GREEN_STRATEGY) {
    $Env:current_cluster_name = $vmClusterResNames[1]
    $cachedLastDeployedCluster = $null
    try {
      # restore_cache_files throws an error on ps1 if nothing is found
      execute_command "restore_cache_files ${lastDeployedClusterFilename} ${stepStateDir}"

      $cachedLastDeployedCluster = Get-Content "${stepStateDir}/${lastDeployedClusterFilename}"
      Remove-Item "${stepStateDir}/${lastDeployedClusterFilename}"
    } catch {
      # no file was found or it was blank.
      $LASTEXITCODE = 0
    }
    # If we have cached the name of the last deployed cluster, deploy to the other one.
    if ($null -ne $cachedLastDeployedCluster -and $cachedLastDeployedCluster -eq $Env:target_cluster_name) {
      $Env:current_cluster_name = $cachedLastDeployedCluster
      $Env:target_cluster_name = $vmClusterResNames[1]
      execute_command "Write-Output 'Previous deployment was to cluster: $cachedLastDeployedCluster. Deploying to cluster: $Env:target_cluster_name'"
    }

  }

  $vmTargets = (Get-Variable -Name "res_$( $Env:target_cluster_name )_targets").Value | ConvertFrom-Json
  $sshOptions = "${baseSSHOptions} -i ~/.ssh/${Env:target_cluster_name}"

  if ($null -ne $deploy_targets_override -and $strategy -ne $BLUE_GREEN_STRATEGY) {
    execute_command "Write-Output 'Overriding vm deploy targets with: $deploy_targets_override'"
    $vmTargets = $deploy_targets_override.Split(",")
  }

  $Script:deployFailed = $false
  for ($i = 0; $i -lt $vmTargets.Length; $i++) {
    $vmTarget = Invoke-Expression $vmTargets[$i]

    # Wait between deploys if delay was specified
    if ($rolloutDelayCommandConfigVal -ne "" -and $i -ne 0) {
      execute_command "Start-Sleep -Seconds $rolloutDelayCommandConfigVal"
    }

    $sshBaseCmd = "ssh -p ${SSH_PORT} ${sshOptions} ${sshUserConfigVal}@${vmTarget}"

    $makeTargetDirCommand = "${sshBaseCmd} `"mkdir -p ${Script:targetDir}`""

    # Command to upload app tarball to vm
    $uploadCommand = "scp -P ${SSH_PORT} ${sshOptions} .\${tarballName} ${sshUserConfigVal}@${vmTarget}`:${Script:targetDir}"

    # Command to run the deploy command from within the uploaded dir
    $untar = "cd ${Script:targetDir}/; tar -xf ${tarballName}; rm -f ${tarballName};"
    $untarCommand = "${sshBaseCmd} `"${untar}`""

    if ($dry_run -eq $true) {
      $uploadCommand = "${sshBaseCmd} `"echo 'dry_run is true. Skipping artifact upload.'`""
      $untarCommand = "${sshBaseCmd} `"echo 'dry_run is true. Skipping artifact decompression.'`""
    }

    try {
      execute_command "Write-Output Creating target dir on vm."
      execute_command $makeTargetDirCommand

      execute_command "Write-Output Uploading artifacts to vm."
      execute_command $uploadCommand

      execute_command "Write-Output Decompressing artifacts archive on vm."
      execute_command $untarCommand
    }
    catch {
      $Script:deployFailed = $true
      # Don't exit on failed commands if fastFail is specified as false
      if ($fastFailConfigVal -eq $false) {
        continue
      }
      break
    }
  }

  exitIfDeployFailed

  $Script:deployedVMs = @()

  if ( $null -ne $step_configuration_scripts_len ) {
    for ($i = 0; $i -lt $step_configuration_scripts_len; $i++) {
      $scriptContext = find_step_configuration_value("scripts_${i}_context")
      $scriptCommandsJSON = find_step_configuration_value("scripts_${i}_commands")
      $scriptCommands = $($scriptCommandsJSON | ConvertFrom-Json) -join "; "
      $scriptName = find_step_configuration_value("scripts_${i}_name")
      if ( $null -eq $scriptName ) {
        $scriptName = $DEFAULT_SCRIPT_NAME
      }
      $Env:current_command = $scriptName
      switch ($scriptContext) {
        $SCRIPT_CONTEXT_BUILD_NODE {
          execute_command "execute_on_build_node ${scriptName} `${scriptCommands}"
          break
        }
        $SCRIPT_CONTEXT_CURRENT_CLUSTER {
          if ( $strategy -ne $BLUE_GREEN_STRATEGY ) {
            throw "Execution of commands in context ${SCRIPT_CONTEXT_CURRENT_CLUSTER} is only supported for ${BLUE_GREEN_STRATEGY} strategy."
          }
          execute_on_current_cluster $scriptName ${scriptCommands}
          break
        }
        $SCRIPT_CONTEXT_TARGET_CLUSTER {
          execute_on_target_cluster $scriptName ${scriptCommands}
          break
        }
        Default {
          throw "Unrecognized script context '${scriptContext}' for script '${scriptName}'."
        }
      }
      $Env:current_command = ""
    }
  }

  # Deploy was successful.

  # Handle success for blueGreen strategy.
  if ( $strategy -eq $BLUE_GREEN_STRATEGY ) {
    execute_command "Write-Output 'Saving last deployed cluster for blueGreen deployment: ${Env:target_cluster_name}'"
    Write-Output $Env:target_cluster_name > "${stepStateDir}/${lastDeployedClusterFilename}"
    execute_command "add_cache_files ${stepStateDir}/${lastDeployedClusterFilename} ${lastDeployedClusterFilename}"
    # Handle success for non-blueGreen strategies.
  } elseif ( $dry_run -ne $true ) {
    # Create rollback archive
    Foreach ($vmTarget IN $Script:deployedVMs) {
      $sshOptions = "${baseSSHOptions} -i ~/.ssh/${Env:target_cluster_name}"
      $sshBaseCmd = "ssh ${sshOptions} -p ${SSH_PORT} ${sshUserConfigVal}@${vmTarget}"
      $rollbackDir = "/home/${sshUserConfigVal}/${step_name}/rollback"
      # Command to copy artifacts into rollback dir.
      $createRollbackArtifacts = "${sshBaseCmd} `"mkdir -p ${rollbackDir}; rm -rf ${rollbackDir}/*; cp -r ${Script:targetDir}/* ${rollbackDir}`""
      execute_command "Write-Output 'Archiving successful deploy for rollback'"
      execute_command "${createRollbackArtifacts}"
    }
  }
}

check_no_verify_ssl

try {
  DeployApplication
}
catch {
  execute_command "Write-Output Something went wrong: `"${_}`""
  execute_command "throw `"${_}`""
}
